import dataclasses
from copy import deepcopy
from functools import wraps

import typing_extensions as typing
from fntypes.co import Error, Nothing, Ok, Option, Result, Some

from telegrinder.modules import logger
from telegrinder.msgspec_utils import msgspec_convert

from .abc import ABCGlobalContext, CtxVar, CtxVariable, GlobalCtxVar

T = typing.TypeVar("T")
F = typing.TypeVar("F", bound=typing.Callable)
CtxValueT = typing.TypeVar("CtxValueT", default=typing.Any)

if typing.TYPE_CHECKING:
    _: typing.TypeAlias = None
else:
    _ = lambda: None


def type_check(value: object, value_type: type[T]) -> typing.TypeGuard[T]:
    if value_type in (typing.Any, object):
        return True
    match msgspec_convert(value, value_type):
        case Ok(v):
            return type(value) is type(v)
        case Error(_):
            return False


def is_dunder(name: str) -> bool:
    return name.startswith("__") and name.endswith("__")


def get_orig_class(obj: T) -> type[T]:
    return getattr(obj, "__orig_class__", obj.__class__)


def root_protection(func: F) -> F:
    if func.__name__ not in ("__setattr__", "__getattr__", "__delattr__"):
        raise RuntimeError(
            "You cannot decorate a {!r} function with this decorator, only "
            "'__setattr__', __getattr__', '__delattr__' methods.".format(
                func.__name__,
            )
        )

    @wraps(func)
    def wrapper(self: "GlobalContext", name: str, /, *args) -> typing.Any:
        if self.is_root_attribute(name) and name in (self.__dict__ | self.__class__.__dict__):
            root_attr = self.get_root_attribute(name).unwrap()
            if all((not root_attr.can_be_rewritten, not root_attr.can_be_read)):
                raise AttributeError(f"Unable to set, get, delete root attribute {name!r}.")
            if func.__name__ == "__setattr__" and not root_attr.can_be_rewritten:
                raise AttributeError(f"Unable to set root attribute {name!r}.")
            if func.__name__ == "__getattr__" and not root_attr.can_be_read:
                raise AttributeError(f"Unable to get root attribute {name!r}.")
            if func.__name__ == "__delattr__":
                raise AttributeError(f"Unable to delete root attribute {name!r}.")

        return func(self, name, *args)  # type: ignore

    return wrapper  # type: ignore


def ctx_var(value: T, *, const: bool = False) -> T:
    """Example:
    ```
    class MyCtx(GlobalContext):
        name: typing.Final[str]
        URL: typing.Final = ctx_var("https://google.com", const=True)

    ctx = MyCtx(name=ctx_var("Alex", const=True))
    ctx.URL  #: 'https://google.com'
    ctx.URL = '...'  #: type checking error & exception 'TypeError'
    ```
    """

    return typing.cast(T, CtxVar(value, const=const))


@dataclasses.dataclass(frozen=True, eq=False)
class RootAttr:
    name: str
    _: dataclasses.KW_ONLY
    can_be_read: bool = dataclasses.field(default=True, kw_only=True)
    can_be_rewritten: bool = dataclasses.field(default=False, kw_only=True)

    def __eq__(self, __value: str) -> bool:
        return self.name == __value


@dataclasses.dataclass(repr=False, frozen=True)
class Storage:
    _storage: dict[str, "GlobalContext"] = dataclasses.field(
        default_factory=lambda: {},
        init=False,
    )

    def __repr__(self) -> str:
        return "<ContextStorage: %s>" % ", ".join("ctx @" + repr(x) for x in self._storage)

    @property
    def storage(self) -> dict[str, "GlobalContext"]:
        return self._storage.copy()

    def set(self, name: str, ctx: "GlobalContext") -> None:
        self._storage.setdefault(name, ctx)

    def get(self, ctx_name: str) -> Option["GlobalContext"]:
        ctx = self._storage.get(ctx_name)
        return Some(ctx) if ctx is not None else Nothing()

    def delete(self, ctx_name: str) -> None:
        assert (
            self._storage.pop(ctx_name, None) is not None
        ), f"Context {ctx_name!r} is not defined in storage."


@typing.dataclass_transform(
    kw_only_default=True,
    order_default=True,
    field_specifiers=(ctx_var,),
)
class GlobalContext(
    ABCGlobalContext, typing.Generic[CtxValueT], dict[str, GlobalCtxVar[CtxValueT]]
):
    """GlobalContext.

    ```
    ctx = GlobalContext()
    ctx["client"] = Client()
    ctx.address = CtxVar("128.0.0.7:8888", const=True)

    def request():
        data = {"user": "root_user", "password": "secret_password"}
        ctx.client.request(ctx.address + "/login", data)
    """

    __ctx_name__: str | None
    __storage__: typing.ClassVar[Storage] = Storage()
    __root_attributes__: typing.ClassVar[tuple[RootAttr, ...]] = (
        RootAttr("__ctx_name__"),
        RootAttr("__root_attributes__"),
        RootAttr("__storage__"),
    )

    def __new__(
        cls,
        ctx_name: str | None = None,
        /,
        **variables: typing.Any | CtxVar[CtxValueT],
    ) -> typing.Self:
        """Create or get from storage a new `GlobalContext` object."""

        if not issubclass(GlobalContext, cls):
            defaults = {}
            for name in cls.__annotations__:
                if name in cls.__dict__ and name not in cls.__root_attributes__:
                    defaults[name] = getattr(cls, name)
                    delattr(cls, name)
                    if isinstance(defaults[name], CtxVar) and defaults[name].const:
                        variables.pop(name, None)

            variables = defaults | variables

        ctx_name = getattr(cls, "__ctx_name__", ctx_name)
        if ctx_name is None:
            ctx = dict.__new__(cls)
        elif ctx_name in cls.__storage__.storage:
            ctx = cls.__storage__.get(ctx_name).unwrap()
        else:
            ctx = dict.__new__(cls, ctx_name)
            cls.__storage__.set(ctx_name, ctx)

        ctx.set_context_variables(variables)
        return ctx  # type: ignore

    def __init__(
        self,
        ctx_name: str | None = None,
        /,
        **variables: CtxValueT | CtxVariable[CtxValueT],
    ):
        """Initialization of `GlobalContext` with passed variables."""

        if not hasattr(self, "__ctx_name__"):
            self.__ctx_name__ = ctx_name

        if variables and not self:
            self.set_context_variables(variables)

    def __repr__(self) -> str:
        return "<{} -> ({})>".format(
            f"{self.__class__.__name__}@{self.ctx_name!r}",
            ", ".join(repr(var) for var in self),
        )

    def __eq__(self, __value: "GlobalContext") -> bool:
        """Returns True if the names of context stores
        that use self and __value instances are equivalent."""

        return isinstance(__value, GlobalContext) and self.__ctx_name__ == __value.__ctx_name__

    def __setitem__(self, __name: str, __value: CtxValueT | CtxVariable[CtxValueT]):
        if is_dunder(__name):
            raise NameError("Cannot set a context variable with dunder name.")
        var = self.get(__name)
        if var and var.unwrap().const:
            raise TypeError(f"Unable to set variable {__name!r}, because it's a constant.")
        dict.__setitem__(self, __name, GlobalCtxVar.collect(__name, __value))

    def __getitem__(self, __name: str) -> CtxValueT:
        return self.get(__name).unwrap().value

    def __delitem__(self, __name: str):
        var = self.get(__name).unwrap()
        if var.const:
            raise TypeError(f"Unable to delete variable {__name!r}, because it's a constant.")
        dict.__delitem__(self, __name)

    @root_protection
    def __setattr__(self, __name: str, __value: CtxValueT | CtxVariable[CtxValueT]):
        """Setting a context variable."""

        if is_dunder(__name):
            return object.__setattr__(self, __name, __value)
        self.__setitem__(__name, __value)

    @root_protection
    def __getattr__(self, __name: str) -> CtxValueT:
        """Getting a context variable."""

        if is_dunder(__name):
            return object.__getattribute__(self, __name)
        return self.__getitem__(__name)

    @root_protection
    def __delattr__(self, __name: str) -> None:
        """Removing a context variable."""

        if is_dunder(__name):
            return object.__delattr__(self, __name)
        self.__delitem__(__name)

    @property
    def ctx_name(self) -> str:
        """Context name."""

        return self.__ctx_name__ or "<Unnamed ctx at %#x>" % id(self)

    @classmethod
    def is_root_attribute(cls, name: str) -> bool:
        """Returns True if exists root attribute
        otherwise False."""

        return name in cls.__root_attributes__

    def set_context_variables(
        self, variables: typing.Mapping[str, CtxValueT | CtxVariable[CtxValueT]]
    ) -> None:
        """Set context variables from mapping."""

        for name, var in variables.items():
            self[name] = var

    def get_root_attribute(self, name: str) -> Option[RootAttr]:
        """Get root attribute by name."""

        if self.is_root_attribute(name):
            for rattr in self.__root_attributes__:
                if rattr.name == name:
                    return Some(rattr)
        return Nothing()

    def items(self) -> list[tuple[str, GlobalCtxVar[CtxValueT]]]:
        """Return context variables as set-like items."""

        return list(dict.items(self))

    def keys(self) -> list[str]:
        """Returns context variable names as keys."""

        return list(dict.keys(self))

    def values(self) -> list[GlobalCtxVar[CtxValueT]]:
        """Returns context variables as values."""

        return list(dict.values(self))

    def update(self, other: typing.Self) -> None:
        """Update context."""

        dict.update(dict(other.items()))

    def copy(self) -> typing.Self:
        """Copy context. Returns copied context without ctx_name."""

        return self.__class__(**self.dict())

    def dict(self) -> dict[str, GlobalCtxVar[CtxValueT]]:
        """Returns context as dict."""

        return {name: deepcopy(var) for name, var in self.items()}

    @typing.overload
    def pop(self, var_name: str) -> Option[GlobalCtxVar[CtxValueT]]: ...

    @typing.overload
    def pop(
        self,
        var_name: str,
        var_value_type: type[T],
    ) -> Option[GlobalCtxVar[T]]: ...

    def pop(self, var_name: str, var_value_type: type[T] = object) -> Option[GlobalCtxVar[T]]:
        """Pop context variable by name."""

        val = self.get(var_name, var_value_type)
        if val:
            del self[var_name]
            return val
        return Nothing()

    @typing.overload
    def get(self, var_name: str) -> Option[GlobalCtxVar[CtxValueT]]: ...

    @typing.overload
    def get(
        self,
        var_name: str,
        var_value_type: type[T],
    ) -> Option[GlobalCtxVar[T]]: ...

    def get(
        self,
        var_name: str,
        var_value_type: type[T] = object,
    ) -> Option[GlobalCtxVar[T]]:
        """Get context variable by name."""

        generic_types = typing.get_args(get_orig_class(self))
        if generic_types and var_value_type is object:
            var_value_type = generic_types[0]
        var = dict.get(self, var_name)
        if var is None:
            return Nothing()
        assert type_check(
            var.value, var_value_type
        ), "Context variable value type of {!r} does not correspond to the expected type {!r}.".format(
            type(var.value).__name__,
            (
                getattr(var_value_type, "__name__")
                if isinstance(var_value_type, type)
                else repr(var_value_type)
            ),
        )
        return Some(var)

    @typing.overload
    def get_value(self, var_name: str) -> Option[CtxValueT]: ...

    @typing.overload
    def get_value(
        self,
        var_name: str,
        var_value_type: type[T],
    ) -> Option[T]: ...

    def get_value(
        self,
        var_name: str,
        var_value_type: type[T] = object,
    ) -> Option[T]:
        """Get context variable value by name."""

        return self.get(var_name, var_value_type).map(lambda var: var.value)

    def rename(self, old_var_name: str, new_var_name: str) -> Result[_, str]:
        """Rename context variable."""

        var = self.get(old_var_name).unwrap()
        if var.const:
            return Error(f"Unable to rename variable {old_var_name!r}, " "because it's a constant.")
        del self[old_var_name]
        self[new_var_name] = var.value
        return Ok(_())

    def clear(self, *, include_consts: bool = False) -> None:
        """Clear context. If `include_consts = True`,
        then the context is completely cleared."""

        if not self:
            return
        if include_consts:
            logger.warning(
                "Constants from the global context {!r} have been cleaned up!",
                self.ctx_name + " at %#x" % id(self),
            )
            return dict.clear(self)

        for name, var in self.dict().items():
            if not var.const:
                del self[name]

    def delete_ctx(self) -> Result[_, str]:
        """Delete context by `ctx_name`."""

        if not self.__ctx_name__:
            return Error("Cannot delete unnamed context.")
        ctx = self.__storage__.get(self.ctx_name).unwrap()
        dict.clear(ctx)
        self.__storage__.delete(self.ctx_name)
        logger.warning(f"Global context {self.ctx_name!r} has been deleted!")
        return Ok(_())


__all__ = (
    "ABCGlobalContext",
    "CtxVar",
    "CtxVariable",
    "GlobalContext",
    "GlobalCtxVar",
    "RootAttr",
    "Storage",
    "ctx_var",
    "get_orig_class",
    "is_dunder",
    "root_protection",
    "type_check",
)



import abc
import typing

from telegrinder.bot.cute_types import InlineQueryCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.bot.rules.adapter import EventAdapter
from telegrinder.types.enums import ChatType

from .markup import Markup, PatternLike, check_string

InlineQuery: typing.TypeAlias = InlineQueryCute


class InlineQueryRule(ABCRule[InlineQuery], abc.ABC):
    adapter = EventAdapter("inline_query", InlineQuery)

    @abc.abstractmethod
    async def check(self, query: InlineQuery, ctx: Context) -> bool:
        pass


class HasLocation(InlineQueryRule):
    async def check(self, query: InlineQuery, ctx: Context) -> bool:
        return bool(query.location)


class InlineQueryChatType(InlineQueryRule):
    def __init__(self, chat_type: ChatType, /) -> None:
        self.chat_type = chat_type

    async def check(self, query: InlineQuery, ctx: Context) -> bool:
        return query.chat_type.map(lambda x: x == self.chat_type).unwrap_or(False)


class InlineQueryText(InlineQueryRule):
    def __init__(self, texts: str | list[str], *, lower_case: bool = False) -> None:
        self.texts = [
            text.lower() if lower_case else text
            for text in ([texts] if isinstance(texts, str) else texts)
        ]
        self.lower_case = lower_case

    async def check(self, query: InlineQuery, ctx: Context) -> bool:
        return (query.query.lower() if self.lower_case else query.query) in self.texts


class InlineQueryMarkup(InlineQueryRule):
    def __init__(self, patterns: PatternLike | list[PatternLike], /) -> None:
        self.patterns = Markup(patterns).patterns

    async def check(self, query: InlineQuery, ctx: Context) -> bool:
        return check_string(self.patterns, query.query, ctx)


__all__ = (
    "HasLocation",
    "InlineQueryChatType",
    "InlineQueryMarkup",
    "InlineQueryRule",
    "InlineQueryText",
)



import typing
from abc import ABC, abstractmethod

from telegrinder.api.abc import ABCAPI
from telegrinder.tools.global_context import ABCGlobalContext
from telegrinder.types import Update


class ABCDispatch(ABC):
    global_context: ABCGlobalContext

    @abstractmethod
    async def feed(self, event: Update, api: ABCAPI) -> bool:
        pass

    @abstractmethod
    def load(self, external: typing.Self) -> None:
        pass


__all__ = ("ABCDispatch",)



import typing
from abc import abstractmethod

from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.dispatch.middleware import ABCMiddleware
from telegrinder.tools.i18n import ABCI18n, I18nEnum

T = typing.TypeVar("T", bound=BaseCute)


class ABCTranslatorMiddleware(ABCMiddleware[T]):
    def __init__(self, i18n: ABCI18n):
        self.i18n = i18n

    @abstractmethod
    async def get_locale(self, event: T) -> str:
        pass

    async def pre(self, event: T, ctx: dict) -> bool:
        ctx[I18nEnum.I18N] = self.i18n.get_translator_by_locale(await self.get_locale(event))
        return True


__all__ = ("ABCTranslatorMiddleware",)



import msgspec
from fntypes.result import Error, Ok, Result

from telegrinder.api.error import APIError
from telegrinder.model import Model


class APIResponse(Model):
    ok: bool = False
    result: msgspec.Raw = msgspec.Raw(b"")
    error_code: int = 0
    description: str = ""

    def to_result(self) -> Result[msgspec.Raw, APIError]:
        if self.ok:
            return Ok(self.result)
        return Error(APIError(self.error_code, self.description))


__all__ = ("APIResponse",)



import asyncio
import dataclasses
import typing

from vbml.patcher import Patcher

from telegrinder.api.abc import ABCAPI
from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules import ABCRule
from telegrinder.modules import logger
from telegrinder.tools.error_handler.error_handler import ErrorHandler
from telegrinder.tools.global_context import TelegrinderCtx
from telegrinder.types import Update

from .abc import ABCDispatch
from .handler import ABCHandler, FuncHandler
from .handler.func import ErrorHandlerT
from .view.box import (
    CallbackQueryViewT,
    ChatJoinRequestViewT,
    ChatMemberViewT,
    InlineQueryViewT,
    MessageViewT,
    RawEventViewT,
    ViewBox,
)

T = typing.TypeVar("T")
R = typing.TypeVar("R")
P = typing.ParamSpec("P")
Handler = typing.Callable[
    typing.Concatenate[T, ...], typing.Coroutine[typing.Any, typing.Any, typing.Any]
]
Event = typing.TypeVar("Event", bound=BaseCute)

DEFAULT_DATACLASS: typing.Final[type[Update]] = Update


@dataclasses.dataclass(repr=False, kw_only=True)
class Dispatch(
    ABCDispatch,
    ViewBox[
        CallbackQueryViewT,
        ChatJoinRequestViewT,
        ChatMemberViewT,
        InlineQueryViewT,
        MessageViewT,
        RawEventViewT,
    ],
):
    global_context: TelegrinderCtx = dataclasses.field(
        init=False,
        default_factory=lambda: TelegrinderCtx(),
    )
    default_handlers: list[ABCHandler] = dataclasses.field(
        init=False,
        default_factory=lambda: [],
    )

    def __repr__(self) -> str:
        return "Dispatch(%s)" % ", ".join(f"{k}={v!r}" for k, v in self.__dict__.items())

    @property
    def patcher(self) -> Patcher:
        """Alias `patcher` to get `vbml.Patcher` from the global context"""
        return self.global_context.vbml_patcher

    @typing.overload
    def handle(
        self,
        *rules: ABCRule[Event],
    ) -> typing.Callable[[Handler[T]], FuncHandler[Event, Handler[T], ErrorHandler]]: ...

    @typing.overload
    def handle(
        self,
        *rules: ABCRule[Event],
        is_blocking: bool = True,
    ) -> typing.Callable[[Handler[T]], FuncHandler[Event, Handler[T], ErrorHandler]]: ...

    @typing.overload
    def handle(
        self,
        *rules: ABCRule[Event],
        dataclass: type[T],
        is_blocking: bool = True,
    ) -> typing.Callable[[Handler[T]], FuncHandler[Event, Handler[T], ErrorHandler]]: ...

    @typing.overload
    def handle(  # type: ignore
        self,
        *rules: ABCRule[Event],
        error_handler: ErrorHandlerT,
        is_blocking: bool = True,
    ) -> typing.Callable[[Handler[T]], FuncHandler[Event, Handler[T], ErrorHandlerT]]: ...

    @typing.overload
    def handle(
        self,
        *rules: ABCRule[Event],
        dataclass: type[T],
        error_handler: ErrorHandlerT,
        is_blocking: bool = True,
    ) -> typing.Callable[[Handler[T]], FuncHandler[Event, Handler[T], ErrorHandlerT]]: ...

    @typing.overload
    def handle(
        self,
        *rules: ABCRule[Event],
        dataclass: type[T] = DEFAULT_DATACLASS,
        error_handler: typing.Literal[None] = None,
        is_blocking: bool = True,
    ) -> typing.Callable[[Handler[T]], FuncHandler[Event, Handler[T], ErrorHandler]]: ...

    def handle(  # type: ignore
        self,
        *rules: ABCRule,
        dataclass: type[typing.Any] = DEFAULT_DATACLASS,
        error_handler: ErrorHandlerT | None = None,
        is_blocking: bool = True,
    ):
        def wrapper(func: typing.Callable):
            handler = FuncHandler(
                func,
                list(rules),
                is_blocking=is_blocking,
                dataclass=dataclass,
                error_handler=error_handler or ErrorHandler(),
            )
            self.default_handlers.append(handler)
            return handler

        return wrapper

    async def feed(self, event: Update, api: ABCAPI) -> bool:
        logger.debug("Processing update (update_id={})", event.update_id)
        loop = asyncio.get_running_loop()
        loop.create_task(self.raw_event.process(event, api))
        
        for view in self.get_views().values():
            if await view.check(event):
                logger.debug(
                    "Update (update_id={}) matched view {!r}",
                    event.update_id,
                    view.__class__.__name__,
                )
                loop.create_task(view.process(event, api))
                return True

        ctx = Context()
        found = False
        for handler in self.default_handlers:
            if await handler.check(api, event, ctx):
                found = True
                loop.create_task(handler.run(event, ctx))
                if handler.is_blocking:
                    break
        return found

    def load(self, external: typing.Self) -> None:
        view_external = external.get_views()
        for name, view in self.get_views().items():
            assert name in view_external, f"View {name!r} is undefined in external dispatch."
            view.load(view_external[name])
            setattr(external, name, view)


__all__ = ("Dispatch",)



import typing

from fntypes.co import Nothing, Some

from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.cute_types.update import UpdateCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.msgspec_utils import Option
from telegrinder.types.enums import ChatType, DiceEmoji
from telegrinder.types.objects import User

from .abc import ABCRule, Message
from .message import MessageRule

T = typing.TypeVar("T", bound=BaseCute)


def get_from_user(obj: typing.Any) -> User:
    assert isinstance(obj, FromUserProto)
    return obj.from_.unwrap() if isinstance(obj.from_, Some | Nothing) else obj.from_


@typing.runtime_checkable
class FromUserProto(typing.Protocol):
    from_: User | Option[User]


class HasFrom(ABCRule[T]):
    async def check(self, event: UpdateCute, ctx: Context) -> bool:
        event_model = event.incoming_update.unwrap()
        return isinstance(event_model, FromUserProto) and bool(event_model.from_)


class HasDice(MessageRule):
    async def check(self, message: Message, ctx: Context) -> bool:
        return bool(message.dice)


class IsForward(MessageRule):
    async def check(self, message: Message, ctx: Context) -> bool:
        return bool(message.forward_origin)


class IsForwardType(MessageRule, requires=[IsForward()]):
    def __init__(
        self, fwd_type: typing.Literal["user", "hidden_user", "chat", "channel"], /
    ) -> None:
        self.fwd_type = fwd_type

    async def check(self, message: Message, ctx: Context) -> bool:
        return message.forward_origin.unwrap().v.type == self.fwd_type


class IsReply(MessageRule):
    async def check(self, message: Message, ctx: Context) -> bool:
        return bool(message.reply_to_message)


class IsSticker(MessageRule):
    async def check(self, message: Message, ctx: Context) -> bool:
        return bool(message.sticker)


class IsBot(ABCRule[T], requires=[HasFrom()]):
    async def check(self, event: UpdateCute, ctx: Context) -> bool:
        return get_from_user(event.incoming_update.unwrap()).is_bot


class IsUser(ABCRule[T], requires=[HasFrom()]):
    async def check(self, event: UpdateCute, ctx: Context) -> bool:
        return not get_from_user(event.incoming_update.unwrap()).is_bot


class IsPremium(ABCRule[T], requires=[HasFrom()]):
    async def check(self, event: UpdateCute, ctx: Context) -> bool:
        return get_from_user(event.incoming_update.unwrap()).is_premium.unwrap_or(False)


class IsLanguageCode(ABCRule[T], requires=[HasFrom()]):
    def __init__(self, lang_codes: str | list[str], /) -> None:
        self.lang_codes = [lang_codes] if isinstance(lang_codes, str) else lang_codes

    async def check(self, event: UpdateCute, ctx: Context) -> bool:
        return (
            get_from_user(event.incoming_update.unwrap()).language_code.unwrap_or_none()
            in self.lang_codes
        )


class IsForum(MessageRule):
    async def check(self, message: Message, ctx: Context) -> bool:
        return message.chat.is_forum.unwrap_or(False)


class IsUserId(ABCRule[T], requires=[HasFrom()]):
    def __init__(self, user_ids: int | list[int], /) -> None:
        self.user_ids = [user_ids] if isinstance(user_ids, int) else user_ids

    async def check(self, event: UpdateCute, ctx: Context) -> bool:
        return get_from_user(event.incoming_update.unwrap()).id in self.user_ids


class IsChatId(MessageRule):
    def __init__(self, chat_ids: int | list[int], /) -> None:
        self.chat_ids = [chat_ids] if isinstance(chat_ids, int) else chat_ids

    async def check(self, message: Message, ctx: Context) -> bool:
        return message.chat.id in self.chat_ids


class IsPrivate(MessageRule):
    async def check(self, message: Message, ctx: Context) -> bool:
        return message.chat.type == ChatType.PRIVATE


class IsGroup(MessageRule):
    async def check(self, message: Message, ctx: Context) -> bool:
        return message.chat.type == ChatType.GROUP


class IsSuperGroup(MessageRule):
    async def check(self, message: Message, ctx: Context) -> bool:
        return message.chat.type == ChatType.SUPERGROUP


class IsChat(MessageRule):
    async def check(self, message: Message, ctx: Context) -> bool:
        return message.chat.type in (ChatType.GROUP, ChatType.SUPERGROUP)


class IsDiceEmoji(MessageRule, requires=[HasDice()]):
    def __init__(self, dice_emoji: DiceEmoji, /) -> None:
        self.dice_emoji = dice_emoji

    async def check(self, message: Message, ctx: Context) -> bool:
        return message.dice.unwrap().emoji == self.dice_emoji


__all__ = (
    "IsBot",
    "IsChat",
    "IsChatId",
    "IsDiceEmoji",
    "IsForum",
    "IsForward",
    "IsForwardType",
    "IsGroup",
    "IsLanguageCode",
    "IsPremium",
    "IsPrivate",
    "IsReply",
    "IsSuperGroup",
    "IsUser",
    "IsUserId",
)



import abc
import typing

from fntypes.result import Result

from telegrinder.api.abc import ABCAPI
from telegrinder.bot.cute_types import BaseCute
from telegrinder.bot.rules.adapter.errors import AdapterError
from telegrinder.model import Model

UpdateT = typing.TypeVar("UpdateT", bound=Model)
CuteT = typing.TypeVar("CuteT", bound=BaseCute)


class ABCAdapter(abc.ABC, typing.Generic[UpdateT, CuteT]):
    @abc.abstractmethod
    async def adapt(self, api: ABCAPI, update: UpdateT) -> Result[CuteT, AdapterError]:
        pass


__all__ = ("ABCAdapter",)



import typing

from telegrinder.api.abc import ABCAPI
from telegrinder.bot.cute_types import MessageCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.process import check_rule
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.modules import logger
from telegrinder.msgspec_utils import Nothing
from telegrinder.types.objects import ReplyParameters, Update

from .abc import ABCHandler


class MessageReplyHandler(ABCHandler[MessageCute]):
    def __init__(
        self,
        text: str,
        *rules: ABCRule[MessageCute],
        is_blocking: bool = True,
        as_reply: bool = False,
    ) -> None:
        self.text = text
        self.rules = list(rules)
        self.as_reply = as_reply
        self.is_blocking = is_blocking
        self.preset_context = Context()

    def __repr__(self) -> str:
        return "<{}: with rules={!r}, {}: {!r}>".format(
            ("blocking " if self.is_blocking else "") + self.__class__.__name__,
            self.rules,
            "answer text as reply" if self.as_reply else "answer text",
            self.text,
        )

    async def check(self, api: ABCAPI, event: Update, ctx: Context | None = None) -> bool:
        ctx = ctx or Context()
        temp_ctx = ctx.copy()
        temp_ctx |= self.preset_context

        for rule in self.rules:
            if not await check_rule(api, rule, event, ctx):
                logger.debug("Rule {!r} failed!", rule)
                return False

        ctx |= temp_ctx
        return True

    async def run(self, event: MessageCute, _: Context) -> typing.Any:
        await event.answer(
            text=self.text,
            reply_parameters=ReplyParameters(event.message_id) if self.as_reply else None,
        )


__all__ = ("MessageReplyHandler",)



from telegrinder.bot.dispatch.context import Context
from telegrinder.types.enums import MessageEntityType

from .text import Message, TextMessageRule


class HasMention(TextMessageRule):
    async def check(self, message: Message, ctx: Context) -> bool:
        if not message.entities.unwrap_or_none():
            return False
        return any(entity.type == MessageEntityType.MENTION for entity in message.entities.unwrap())


__all__ = ("HasMention",)



import typing

from telegrinder.bot.cute_types import UpdateCute
from telegrinder.node import Node


class NodeSession:
    def __init__(
        self,
        value: typing.Any,
        subnodes: dict[str, typing.Self],
        generator: typing.AsyncGenerator[typing.Any, None] | None = None,
    ):
        self.value = value
        self.subnodes = subnodes
        self.generator = generator

    async def close(self, with_value: typing.Any | None = None) -> None:
        for subnode in self.subnodes.values():
            await subnode.close()

        if self.generator is None:
            return
        try:
            await self.generator.asend(with_value)
        except StopAsyncIteration:
            self.generator = None

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.value}" + ("ACTIVE>" if self.generator else ">")


class NodeCollection:
    def __init__(self, sessions: dict[str, NodeSession]) -> None:
        self.sessions = sessions

    def values(self) -> dict[str, typing.Any]:
        return {name: session.value for name, session in self.sessions.items()}

    async def close_all(self, with_value: typing.Any | None = None) -> None:
        for session in self.sessions.values():
            await session.close(with_value)


async def compose_node(
    node: type[Node],
    update: UpdateCute,
    ready_context: dict[str, NodeSession] | None = None,
) -> NodeSession:
    _node = node.as_node()
    context = NodeCollection(ready_context.copy() if ready_context else {})

    for name, subnode in _node.get_sub_nodes().items():
        if subnode is UpdateCute:
            context.sessions[name] = NodeSession(update, {})
        else:
            context.sessions[name] = await compose_node(subnode, update)

    generator: typing.AsyncGenerator | None

    if _node.is_generator():
        generator = typing.cast(typing.AsyncGenerator, _node.compose(**context.values()))
        value = await generator.asend(None)
    else:
        generator = None
        value = await _node.compose(**context.values())  # type: ignore

    return NodeSession(value, context.sessions, generator)


__all__ = ("NodeCollection", "NodeSession", "compose_node")



import typing

import vbml

from telegrinder.tools.global_context import GlobalContext, ctx_var


class TelegrinderCtx(GlobalContext):
    """Basic type-hinted telegrinder context with context name `"telegrinder"`.

    You can use this class or GlobalContext:
    ```
    from telegrinder.tools.global_context import GlobalContext, TelegrinderCtx

    ctx1 = TelegrinderCtx()
    ctx2 = GlobalContext("telegrinder")  # same, but without the type-hints
    assert ctx1 == ctx2  # ok
    ```"""

    __ctx_name__ = "telegrinder"

    vbml_patcher: typing.ClassVar[vbml.Patcher] = ctx_var(vbml.Patcher(), const=True)


__all__ = ("TelegrinderCtx",)



import os
import re
import typing

import yaml

from telegrinder.tools.keyboard import InlineKeyboard, Keyboard

from .base import KeyboardSetBase, KeyboardSetError

PathLike: typing.TypeAlias = str | os.PathLike[str]


class KeyboardSetYAML(KeyboardSetBase):
    __config__: PathLike

    @classmethod
    def load(cls) -> None:
        config_path = getattr(cls, "__config__", "keyboards.yaml")
        if not os.path.exists(config_path):
            raise FileNotFoundError(f"Config file for {cls.__name__!r} is undefined.")

        config = yaml.load(
            open(  # noqa: SIM115
                str(config_path),
                mode="r",
                encoding="UTF-8",
            ),
            yaml.Loader,
        )
        for name, hint in typing.get_type_hints(cls).items():
            g = re.match(r"(?:kb_|keyboard_)(.+)", name.lower())
            if not g:
                continue

            short_name = g.group(1)
            if short_name not in config:
                raise KeyboardSetError(f"Keyboard {short_name!r} is undefined in config.")

            kb_config = config[short_name]
            if (
                not isinstance(kb_config, dict)
                or "buttons" not in kb_config
                or not isinstance(kb_config["buttons"], list)
            ):
                raise KeyboardSetError(
                    "Keyboard should be dict with field buttons which must be a list, "
                    "check documentation."
                )

            buttons = kb_config.pop("buttons")
            new_keyboard: Keyboard | InlineKeyboard = hint(**kb_config)
            for button in buttons:
                if not button:
                    new_keyboard.row()
                    continue
                if "text" not in button:
                    raise KeyboardSetError("Text is required in button.")
                new_keyboard.add(new_keyboard.BUTTON(**button))  # type: ignore

            setattr(cls, name, new_keyboard)


__all__ = ("KeyboardSetYAML",)



import typing

from telegrinder.bot.cute_types import InlineQueryCute
from telegrinder.bot.dispatch.context import Context

from .abc import BaseReturnManager, register_manager


class InlineQueryReturnManager(BaseReturnManager[InlineQueryCute]):
    @register_manager(dict)
    @staticmethod
    async def dict_manager(
        value: dict[str, typing.Any], event: InlineQueryCute, ctx: Context
    ) -> None:
        await event.answer(**value)


__all__ = ("InlineQueryReturnManager",)
